"""Label queries."""

from typing import Dict, Generator, Iterable, List, Optional, cast, overload

import pandas as pd
from typeguard import typechecked
from typing_extensions import Literal

from kili import services
from kili.authentication import KiliAuth
from kili.graphql import QueryOptions
from kili.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.graphql.operations.label.queries import LabelQuery, LabelWhere
from kili.helpers import disable_tqdm_if_as_generator, validate_category_search_query
from kili.services.export.exceptions import NoCompatibleJobError
from kili.services.export.types import CocoAnnotationModifier, LabelFormat, SplitOption
from kili.services.helpers import infer_ids_from_external_ids
from kili.services.types import ProjectId
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class QueriesLabel:
    """Set of Label queries."""

    # pylint: disable=too-many-arguments,too-many-locals,dangerous-default-value

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: List[str] = [
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
        ],
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: List[str] = [
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
        ],
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: List[str] = [
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
        ],
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a label list or a label generator from a project based on a set of criteria.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have their creation date equal to this date.
            created_at_gte: Returned labels should have their creation date greater or equal to this date.
            created_at_lte: Returned labels should have their creation date lower or equal to this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.labels(project_id=project_id, fields=['jsonResponse', 'labelOf.externalId']) # returns a list of all labels of a project and their assets external ID
            >>> kili.labels(project_id=project_id, fields=['jsonResponse'], as_generator=True) # returns a generator of all labels of a project

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

            Example:

                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                category_search = `(JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """

        if category_search:
            validate_category_search_query(category_search)

        where = LabelWhere(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            type_in=type_in,
            user_id=user_id,
            category_search=category_search,
        )
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        labels_gen = LabelQuery(self.auth.client, self.auth.ssl_verify)(where, fields, options)

        if as_generator:
            return labels_gen
        return list(labels_gen)

    @overload
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def predictions(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get prediction labels from a project based on a set of criteria.

        This method is equivalent to the `kili.labels()` method, but it only returns label of type "PREDICTION".

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.predictions(project_id=project_id) # returns a list of prediction labels of a project
        """
        if fields is None:
            fields = [
                "author.email",
                "author.id",
                "id",
                "jsonResponse",
                "labelType",
                "modelName",
            ]
        return self.labels(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            fields=fields,
            first=first,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            skip=skip,
            type_in=["PREDICTION"],
            user_id=user_id,
            disable_tqdm=disable_tqdm,
            category_search=category_search,
            as_generator=as_generator,  # type: ignore
        )

    @overload
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def inferences(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: Optional[List[str]] = None,
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        category_search: Optional[str] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get inference labels from a project based on a set of criteria.

        This method is equivalent to the `kili.labels()` method, but it only returns label of type "INFERENCE".

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED`, `TO REVIEW` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have been made by authors in that list, if given.
                An author can be designated by the first name, the last name, or the first name + last name.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.
            category_search: Query to filter labels based on the content of their jsonResponse

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.inferences(project_id=project_id) # returns a list of inference labels of a project
        """
        if fields is None:
            fields = [
                "author.email",
                "author.id",
                "id",
                "jsonResponse",
                "labelType",
                "modelName",
            ]
        return self.labels(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            fields=fields,
            first=first,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            skip=skip,
            type_in=["INFERENCE"],
            user_id=user_id,
            disable_tqdm=disable_tqdm,
            category_search=category_search,
            as_generator=as_generator,  # type: ignore
        )

    @typechecked
    def export_labels_as_df(
        self,
        project_id: str,
        fields: List[str] = [
            "author.email",
            "author.id",
            "createdAt",
            "id",
            "labelType",
        ],
        asset_fields: List[str] = ["externalId"],
    ) -> pd.DataFrame:
        # pylint: disable=line-too-long
        """Get the labels of a project as a pandas DataFrame.

        Args:
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
            asset_fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.

        Returns:
            pandas DataFrame containing the labels.
        """

        services.get_project(self.auth, project_id, ["id"])
        assets_gen = AssetQuery(self.auth.client, self.auth.ssl_verify)(
            AssetWhere(project_id=project_id),
            asset_fields + ["labels." + field for field in fields],
            QueryOptions(disable_tqdm=False),
        )
        labels = [
            dict(
                label,
                **dict((f"asset_{key}", asset[key]) for key in asset if key != "labels"),
            )
            for asset in assets_gen
            for label in asset["labels"]
        ]
        labels_df = pd.DataFrame(labels)
        return labels_df

    @typechecked
    def count_labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        label_id: Optional[str] = None,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        category_search: Optional[str] = None,
        id_contains: Optional[List[str]] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Get the number of labels for the given parameters.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            label_id: Identifier of the label.
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            category_search: Query to filter labels based on the content of their jsonResponse
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of labels with the parameters provided
        """

        if category_search:
            validate_category_search_query(category_search)

        where = LabelWhere(
            project_id=project_id,
            asset_id=asset_id,
            asset_status_in=asset_status_in,
            asset_external_id_in=asset_external_id_in,
            author_in=author_in,
            created_at=created_at,
            created_at_gte=created_at_gte,
            created_at_lte=created_at_lte,
            honeypot_mark_gte=honeypot_mark_gte,
            honeypot_mark_lte=honeypot_mark_lte,
            id_contains=id_contains,
            label_id=label_id,
            type_in=type_in,
            user_id=user_id,
            category_search=category_search,
        )
        return LabelQuery(self.auth.client, self.auth.ssl_verify).count(where)

    def export_labels(
        self,
        project_id: str,
        filename: str,
        fmt: LabelFormat,
        asset_ids: Optional[List[str]] = None,
        layout: SplitOption = "split",
        single_file: bool = False,
        disable_tqdm: bool = False,
        with_assets: bool = True,
        external_ids: Optional[List[str]] = None,
        annotation_modifier: Optional[CocoAnnotationModifier] = None,
    ):
        """
        Export the project labels with the requested format into the requested output path.

        Args:
            project_id: Identifier of the project.
            filename: Relative or full path of the archive that will contain
                the exported data.
            fmt: Format of the exported labels.
            asset_ids: Optional list of the assets internal IDs from which to export the labels.
            layout: Layout of the exported files. "split" means there is one folder
                per job, "merged" that there is one folder with every labels.
            single_file: Layout of the exported labels. Single file mode is
                only available for some specific formats (COCO and Kili).
            disable_tqdm: Disable the progress bar if True.
            with_assets: Download the assets in the export.
            external_ids: Optional list of the assets external IDs from which to export the labels.
            annotation_modifier: For COCO export only: function that takes the COCO annotation, the
                COCO image, and the Kili annotation, and should return an updated COCO annotation.
                This can be used if you want to add a new attribute to the COCO annotation. For
                example, you can add a method that computes if the annotation is a rectangle or not
                and add it to the COCO annotation (see example).

        !!! warning
            Export is not allowed for projects connected to a cloud storage.

        !!! Info
            The supported formats are:

            - Yolo V4, V5, V7 for object detection tasks (bounding box)

            - Kili for all tasks.

            - COCO for semantic segmentation tasks (bounding box and semantic segmentation)

            - Pascal VOC for object detection tasks.

        !!! Example
            ```
            from kili.client import Kili
            kili = Kili()
            kili.export_labels("your_project_id", "export.zip", "yolo_v4")
            ```

        !!! Example
            ```
            from kili.client import Kili
            kili = Kili()

            def is_rectangle(coco_annotation, coco_image, kili_annotation):
                is_rectangle = ...
                return {**coco_annotation, "attributes": {"is_rectangle": is_rectangle}}
            kili.export_labels(
                "your_project_id",
                "export.zip",
                "coco",
                annotation_modifier = add_is_rectangle
            )
            ```
        """
        if external_ids is not None and asset_ids is None:
            id_map = infer_ids_from_external_ids(
                auth=self.auth, asset_external_ids=external_ids, project_id=project_id
            )
            asset_ids = [id_map[id] for id in external_ids]

        try:
            services.export_labels(
                self.auth,
                asset_ids=asset_ids,
                project_id=cast(ProjectId, project_id),
                export_type="latest",
                label_format=fmt,
                split_option=layout,
                single_file=single_file,
                output_file=filename,
                disable_tqdm=disable_tqdm,
                log_level="WARNING",
                with_assets=with_assets,
                annotation_modifier=annotation_modifier,
            )
        except NoCompatibleJobError as excp:
            print(str(excp))



"""
This script permits to initialize the Kili Python SDK client.
"""
import os
from typing import Union

from kili.authentication import KiliAuth
from kili.exceptions import AuthenticationFailed
from kili.graphql.graphql_client import GraphQLClientName
from kili.internal import KiliInternal
from kili.mutations.asset import MutationsAsset
from kili.mutations.data_connection import MutationsDataConnection
from kili.mutations.issue import MutationsIssue
from kili.mutations.label import MutationsLabel
from kili.mutations.notification import MutationsNotification
from kili.mutations.plugins import MutationsPlugins
from kili.mutations.project import MutationsProject
from kili.mutations.project_version import MutationsProjectVersion
from kili.mutations.user import MutationsUser
from kili.project import Project
from kili.queries.api_key import QueriesApiKey
from kili.queries.asset import QueriesAsset
from kili.queries.data_connection import QueriesDataConnection
from kili.queries.data_integration import QueriesDataIntegration
from kili.queries.issue import QueriesIssue
from kili.queries.label import QueriesLabel
from kili.queries.notification import QueriesNotification
from kili.queries.organization import QueriesOrganization
from kili.queries.plugins import QueriesPlugins
from kili.queries.project import QueriesProject
from kili.queries.project_user import QueriesProjectUser
from kili.queries.project_version import QueriesProjectVersion
from kili.queries.user import QueriesUser
from kili.services.project import get_project
from kili.services.types import ProjectId
from kili.subscriptions.label import SubscriptionsLabel

from .helpers import deprecate


class Kili(  # pylint: disable=too-many-ancestors
    MutationsAsset,
    MutationsDataConnection,
    MutationsIssue,
    MutationsLabel,
    MutationsNotification,
    MutationsPlugins,
    MutationsProject,
    MutationsProjectVersion,
    MutationsUser,
    QueriesApiKey,
    QueriesAsset,
    QueriesDataConnection,
    QueriesDataIntegration,
    QueriesIssue,
    QueriesLabel,
    QueriesNotification,
    QueriesOrganization,
    QueriesPlugins,
    QueriesProject,
    QueriesProjectUser,
    QueriesProjectVersion,
    QueriesUser,
    SubscriptionsLabel,
):
    """
    Kili Client.
    """

    def __init__(
        self,
        api_key=None,
        api_endpoint=None,
        verify: Union[bool, str] = True,
        client_name=GraphQLClientName.SDK,
    ):
        """
        Args:
            api_key: User API key generated
                from https://cloud.kili-technology.com/label/my-account/api-key
                Default to  KILI_API_KEY environment variable).
                If not passed, requires the KILI_API_KEY environment variable to be set.
            api_endpoint: Recipient of the HTTP operation
                Default to  KILI_API_ENDPOINT environment variable).
                If not passed, default to Kili SaaS:
                'https://cloud.kili-technology.com/api/label/v2/graphql'
                verify: Verify certificate (see request's documentation). Set to False on
                local deployment without SSL.
            client_name: For internal use only.
                Define the name of the graphQL client whith which graphQL calls will be sent.

        Returns:
            Object container your API session

        Examples:
            list:
                - your assets with: `kili.assets()`
                - your labels with: `kili.labels()`
                - your projects with: `kili.projects()`
        """
        if api_key is None:
            api_key = os.getenv("KILI_API_KEY")
        if api_endpoint is None:
            api_endpoint = os.getenv(
                "KILI_API_ENDPOINT",
                "https://cloud.kili-technology.com/api/label/v2/graphql",
            )

        if api_key is None:
            raise AuthenticationFailed(api_key, api_endpoint)
        try:
            self.auth = KiliAuth(
                api_key=api_key,
                api_endpoint=api_endpoint,
                client_name=client_name,
                ssl_verify=verify,
            )
            super().__init__(self.auth)
        except Exception as exception:
            exception_str = str(exception)
            if "b'Unauthorized'" in exception_str:
                raise AuthenticationFailed(api_key, api_endpoint) from exception
            raise exception

        self.internal = KiliInternal(self)

    @deprecate(
        msg="This method is deprecated. Use `kili.projects(project_id='<MY_PROJECT_ID>')` instead",
        removed_in="2.131",
    )
    def get_project(self, project_id: str) -> Project:
        """Return a project object corresponding to the project_id given.
        The returned project object inherit from many methods for project management

        Args:
            project_id: id of the project to return

        raise:
            NotFound if the given `project_id` does not correspond to an existing project
        """
        project = get_project(self.auth, project_id, ["inputType", "title"])
        return Project(
            client=self,
            project_id=ProjectId(project_id),
            input_type=project["inputType"],
            title=project["title"],
        )



"""
Common code for the coco exporter.
"""

import json
import time
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from kili.orm import Asset, JobMLTask, JobTool
from kili.services.export.exceptions import NoCompatibleJobError, NotCompatibleInputType
from kili.services.export.format.base import AbstractExporter
from kili.services.export.format.coco.types import (
    CocoAnnotation,
    CocoCategory,
    CocoFormat,
    CocoImage,
)
from kili.services.export.types import CocoAnnotationModifier
from kili.services.types import Job, JobName
from kili.utils.tqdm import tqdm

from ...media.image import get_image_dimensions
from ...media.video import cut_video, get_video_dimensions

DATA_SUBDIR = "data"


# COCO format


class CocoExporter(AbstractExporter):
    """
    Common code for COCO exporter.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.with_assets:
            warnings.warn(
                "For an export to the COCO format, the download of assets cannot be disabled.",
                stacklevel=2,
            )
        self.with_assets = True

    def _check_arguments_compatibility(self):
        """
        Checks if the export label format is compatible with the export options.
        """

    def _check_project_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the project.
        """
        if self.project_input_type not in ("IMAGE", "VIDEO"):
            raise NotCompatibleInputType(
                f"Project with input type '{self.project_input_type}' not compatible with COCO"
                " export format."
            )

        if len(self.compatible_jobs) == 0:
            raise NoCompatibleJobError(
                f"Project needs at least one {JobMLTask.ObjectDetection} task with bounding boxes"
                " or segmentations."
            )

    @property
    def images_folder(self) -> Path:
        """
        Export images folder
        """
        return self.base_folder / DATA_SUBDIR

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """
        Extract formatted annotations from labels.
        """
        clean_assets = self.pre_process_assets(assets, self.label_format)

        self._save_assets_export(
            clean_assets, self.export_root_folder, annotation_modifier=self.annotation_modifier
        )
        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _save_assets_export(
        self,
        assets: List[Asset],
        output_directory: Path,
        annotation_modifier: Optional[CocoAnnotationModifier],
    ):
        """
        Save the assets to a file and return the link to that file
        """
        if self.split_option == "split":
            for job_name, job in self.project_json_interface["jobs"].items():
                if self._is_job_compatible(job):
                    _convert_kili_semantic_to_coco(
                        jobs={job_name: job},
                        assets=assets,
                        output_dir=Path(output_directory) / self.project_id,
                        title=self.project_title,
                        project_input_type=self.project_input_type,
                        annotation_modifier=annotation_modifier,
                        merged=False,
                    )
                else:
                    self.logger.warning(f"Job {job_name} is not compatible with the COCO format.")
        else:  # merged
            _convert_kili_semantic_to_coco(
                jobs={
                    k: job
                    for k, job in self.project_json_interface["jobs"].items()
                    if self._is_job_compatible(job)
                },
                assets=assets,
                output_dir=Path(output_directory) / self.project_id,
                title=self.project_title,
                project_input_type=self.project_input_type,
                annotation_modifier=annotation_modifier,
                merged=True,
            )

    def _is_job_compatible(self, job: Job) -> bool:
        if "tools" not in job:
            return False
        return (JobTool.Semantic in job["tools"] or JobTool.Rectangle in job["tools"]) and job[
            "mlTask"
        ] == JobMLTask.ObjectDetection


# pylint: disable=too-many-arguments
def _convert_kili_semantic_to_coco(
    jobs: Dict[JobName, Job],
    assets: List[Asset],
    output_dir: Path,
    title: str,
    project_input_type: str,
    annotation_modifier: Optional[CocoAnnotationModifier],
    merged: bool,
) -> Tuple[CocoFormat, List[Path]]:
    """
    creates the following structure on the disk:
    <dataset_dir>/
        data/
            <filename0>.<ext>
            <filename1>.<ext>
            ...
        labels.json


    We iterate on the assets and create a coco format for each asset.

    Note: the jobs should only contains elligible jobs.
    """

    infos_coco = {
        "year": time.strftime("%Y"),
        "version": "1.0",
        "description": f"{title} - Exported from Kili Python Client",
        "contributor": "Kili Technology",
        "url": "https://kili-technology.com",
        "date_created": datetime.now().isoformat(),
    }
    labels_json = CocoFormat(
        info=infos_coco,
        licenses=[],
        categories=[],
        images=[],
        annotations=[],
    )

    # Mapping category - category id
    cat_kili_id_to_coco_id, labels_json["categories"] = _get_coco_categories_with_mapping(
        jobs, merged
    )

    if merged:
        labels_json["images"], labels_json["annotations"] = _get_coco_images_and_annotations(
            jobs,
            assets,
            cat_kili_id_to_coco_id,
            project_input_type,
            annotation_modifier,
            is_single_job=False,
        )

        label_file_name = output_dir / "labels.json"
        label_file_name.parent.mkdir(parents=True, exist_ok=True)
        with label_file_name.open("w") as outfile:
            json.dump(labels_json, outfile)
        label_filenames = [label_file_name]
    else:  # split case
        label_filenames = []
        for job_name, job in jobs.items():
            labels_json["images"], labels_json["annotations"] = _get_coco_images_and_annotations(
                {job_name: job},
                assets,
                cat_kili_id_to_coco_id,
                project_input_type,
                annotation_modifier,
                is_single_job=True,
            )
            label_file_name = output_dir / job_name / "labels.json"

            label_file_name.parent.mkdir(parents=True, exist_ok=True)
            with label_file_name.open("w") as outfile:
                json.dump(labels_json, outfile)
            label_filenames.append(label_file_name)

    return labels_json, label_filenames


def _get_coco_categories_with_mapping(
    jobs: Dict[JobName, Job], merged: bool
) -> Tuple[Dict[JobName, Dict[str, int]], List[CocoCategory]]:
    """
    Get the mapping between a category name in Kili of a given job and the COCO category id, and
    also return the list of COCO categories.
    """
    if merged:
        # id_offset: int = 0
        mapping_cat_name_cat_kili_id: Dict[str, str] = {}
        cat_kili_id_to_coco_id: Dict[JobName, Dict[str, int]] = {}
        id_offset = 0
        for job_name, job in sorted(jobs.items(), key=lambda x: x[0]):
            content = dict(job["content"])
            job_cats: Dict[str, Any] = content["categories"]
            mapping_cat_name_cat_kili_id = {
                "/".join([job_name, cat["name"]]): cat_id for cat_id, cat in job_cats.items()
            }

            cat_kili_ids = list(sorted(mapping_cat_name_cat_kili_id.values()))
            cat_kili_id_to_coco_id[job_name] = {
                str(category_id): i + id_offset for i, category_id in enumerate(cat_kili_ids)
            }
            id_offset += len(cat_kili_ids)

    else:
        assert (
            len(list(jobs.values())) == 1
        ), "When this method is called with merged = False, the jobs should only contain 1 job"
        job_name = list(jobs.keys())[0]
        cats = list(jobs.values())[0]["content"]["categories"]
        mapping_cat_name_cat_kili_id = {cat["name"]: cat_id for cat_id, cat in cats.items()}
        cat_kili_ids = list(mapping_cat_name_cat_kili_id.values())
        cat_kili_id_to_coco_id = {
            job_name: {str(category_id): i for i, category_id in enumerate(cat_kili_ids)}
        }

    return cat_kili_id_to_coco_id, _get_coco_categories(cat_kili_id_to_coco_id, merged)


# pylint: disable=too-many-locals
def _get_coco_images_and_annotations(
    jobs: Dict[JobName, Job],
    assets: List[Asset],
    cat_kili_id_to_coco_id: Dict[JobName, Dict[str, int]],
    project_input_type: str,
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    if project_input_type == "IMAGE":
        return _get_images_and_annotation_for_images(
            jobs, assets, cat_kili_id_to_coco_id, annotation_modifier, is_single_job
        )

    if project_input_type == "VIDEO":
        return _get_images_and_annotation_for_videos(
            jobs, assets, cat_kili_id_to_coco_id, annotation_modifier, is_single_job
        )

    raise NotImplementedError(f"No conversion to COCO possible for input type {project_input_type}")


def _get_images_and_annotation_for_images(
    jobs: Dict[JobName, Job],
    assets: List[Asset],
    cat_kili_id_to_coco_id: Dict[JobName, Dict[str, int]],
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    coco_images = []
    coco_annotations = []
    annotation_offset = 0

    for asset_i, asset in tqdm(enumerate(assets), desc="Convert to coco format"):
        width, height = get_image_dimensions(asset["content"])
        coco_image = CocoImage(
            id=asset_i,
            license=0,
            file_name=str(DATA_SUBDIR + "/" + Path(asset["content"]).name),
            height=height,
            width=width,
            date_captured=None,
        )

        coco_images.append(coco_image)
        if is_single_job:
            assert len(list(jobs.keys())) == 1
            job_name = list(jobs.keys())[0]

            if job_name not in asset["latestLabel"]["jsonResponse"]:
                coco_img_annotations = []
                # annotation offset is unchanged
            else:
                coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                    asset["latestLabel"]["jsonResponse"][job_name]["annotations"],
                    cat_kili_id_to_coco_id[job_name],
                    annotation_offset,
                    coco_image,
                    annotation_modifier=annotation_modifier,
                )
            coco_annotations.extend(coco_img_annotations)
        else:
            # coco_img_annotations = []
            for job_name in jobs.keys():
                if job_name not in asset["latestLabel"]["jsonResponse"]:
                    continue
                    # annotation offset is unchanged

                coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                    asset["latestLabel"]["jsonResponse"][job_name]["annotations"],
                    cat_kili_id_to_coco_id[job_name],
                    annotation_offset,
                    coco_image,
                    annotation_modifier=annotation_modifier,
                )
                coco_annotations.extend(coco_img_annotations)
    return coco_images, coco_annotations


def _get_images_and_annotation_for_videos(
    jobs: Dict[JobName, Job],
    assets: List[Asset],
    cat_kili_id_to_coco_id: Dict[JobName, Dict[str, int]],
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    coco_images = []
    coco_annotations = []
    annotation_offset = 0

    for asset in tqdm(assets, desc="Convert to coco format"):
        nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
        leading_zeros = len(str(nbr_frames))

        width = height = 0
        frame_ext = ""
        # jsonContent with frames
        if isinstance(asset["jsonContent"], list) and Path(asset["jsonContent"][0]).is_file():
            width, height = get_image_dimensions(asset["jsonContent"][0])
            frame_ext = Path(asset["jsonContent"][0]).suffix

        # video with shouldUseNativeVideo set to True (no frames available)
        elif Path(asset["content"]).is_file():
            width, height = get_video_dimensions(asset["content"])
            cut_video(asset["content"], asset, leading_zeros, Path(asset["content"]).parent)
            frame_ext = ".jpg"

        else:
            raise FileNotFoundError(f"Could not find frames or video for asset {asset}")

        for frame_i, (frame_id, json_response) in enumerate(
            asset["latestLabel"]["jsonResponse"].items()
        ):
            frame_name = f'{asset["externalId"]}_{str(int(frame_id)+1).zfill(leading_zeros)}'
            coco_image = CocoImage(
                id=frame_i + len(assets),  # add offset to avoid duplicate ids
                license=0,
                file_name=str(DATA_SUBDIR + "/" + f"{frame_name}{frame_ext}"),
                height=height,
                width=width,
                date_captured=None,
            )
            coco_images.append(coco_image)

            if is_single_job:
                job_name = list(jobs.keys())[0]
                if job_name not in json_response:
                    coco_img_annotations = []
                    # annotation offset is unchanged
                else:
                    coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                        json_response[job_name]["annotations"],
                        cat_kili_id_to_coco_id[job_name],
                        annotation_offset,
                        coco_image,
                        annotation_modifier,
                    )
                coco_annotations.extend(coco_img_annotations)
            else:
                for job_name in jobs.keys():
                    if job_name not in asset["latestLabel"]["jsonResponse"]:
                        continue
                    coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                        json_response[job_name]["annotations"],
                        cat_kili_id_to_coco_id[job_name],
                        annotation_offset,
                        coco_image,
                        annotation_modifier,
                    )

                    coco_annotations.extend(coco_img_annotations)

    return coco_images, coco_annotations


def _get_coco_image_annotations(
    annotations_: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, int],
    annotation_offset: int,
    coco_image: CocoImage,
    annotation_modifier: Optional[CocoAnnotationModifier],
) -> Tuple[List[CocoAnnotation], int]:
    coco_annotations = []

    annotation_j = annotation_offset

    for annotation in annotations_:  # we do not use enumerate as some annotations may be empty
        annotation_j += 1

        if not annotation:
            print("continue")
            continue
        bounding_poly = annotation["boundingPoly"]
        bbox, poly = _get_coco_geometry_from_kili_bpoly(
            bounding_poly, coco_image["width"], coco_image["height"]
        )
        if len(poly) < 6:  # twice the number of vertices
            print("A polygon must contain more than 2 points. Skipping this polygon...")
            continue

        categories = annotation["categories"]
        coco_annotation = CocoAnnotation(
            id=annotation_j,
            image_id=coco_image["id"],
            category_id=cat_kili_id_to_coco_id[categories[0]["name"]],
            bbox=bbox,
            # Objects have only one connected part.
            # But a type of object can appear several times on the same image.
            # The limitation of the single connected part comes from Kili.
            segmentation=[poly],
            area=coco_image["height"] * coco_image["width"],
            iscrowd=0,
        )

        if annotation_modifier:
            coco_annotation = annotation_modifier(
                dict(coco_annotation), dict(coco_image), annotation
            )

        coco_annotations.append(coco_annotation)
    return coco_annotations, annotation_j


def _get_coco_geometry_from_kili_bpoly(
    bounding_poly: List[Dict], asset_width: int, asset_height: int
):
    normalized_vertices = bounding_poly[0]["normalizedVertices"]
    p_x = [float(vertice["x"]) * asset_width for vertice in normalized_vertices]
    p_y = [float(vertice["y"]) * asset_height for vertice in normalized_vertices]
    poly_vertices = [(float(x), float(y)) for x, y in zip(p_x, p_y)]
    x_min, y_min = min(p_x), min(p_y)
    x_max, y_max = max(p_x), max(p_y)
    bbox_width, bbox_height = x_max - x_min, y_max - y_min
    bbox = [int(x_min), int(y_min), int(bbox_width), int(bbox_height)]
    poly = [p for vertice in poly_vertices for p in vertice]
    return bbox, poly


def _get_coco_categories(cat_kili_id_to_coco_id, merged) -> List[CocoCategory]:
    categories_coco: List[CocoCategory] = []
    for job_name, mapping in sorted(cat_kili_id_to_coco_id.items(), key=lambda x: x[0]):
        for cat_kili_id, cat_coco_id in sorted(mapping.items(), key=lambda x: x[1]):
            categories_coco.append(
                {
                    "id": cat_coco_id,
                    "name": cat_kili_id if not merged else "/".join([job_name, cat_kili_id]),
                    "supercategory": job_name,
                }
            )

    return categories_coco



"""Service for importing objects into kili"""

from typing import Dict, List

from kili.authentication import KiliAuth
from kili.queries.project import QueriesProject
from kili.services.asset_import.image import ImageDataImporter
from kili.services.asset_import.pdf import PdfDataImporter
from kili.services.asset_import.text import TextDataImporter
from kili.services.asset_import.video import VideoDataImporter
from kili.services.project import get_project_field

from .base import LoggerParams, ProcessingParams, ProjectParams

importer_by_type = {
    "PDF": PdfDataImporter,
    "IMAGE": ImageDataImporter,
    "TEXT": TextDataImporter,
    "VIDEO": VideoDataImporter,
    "VIDEO_LEGACY": VideoDataImporter,
}


def import_assets(  # pylint: disable=too-many-arguments
    auth: KiliAuth,
    project_id: str,
    assets: List[Dict],
    raise_error=True,
    disable_tqdm=False,
    verify=True,
):
    """
    import the selected assets into the specified project
    """
    kili = QueriesProject(auth=auth)
    input_type = get_project_field(kili.auth, project_id, "inputType")

    project_params = ProjectParams(project_id=project_id, input_type=input_type)
    processing_params = ProcessingParams(raise_error=raise_error, verify=verify)
    logger_params = LoggerParams(disable_tqdm=disable_tqdm)
    importer_params = (auth, project_params, processing_params, logger_params)

    if input_type not in importer_by_type:
        raise NotImplementedError(f"There is no imported for the input type: {input_type}")
    asset_importer = importer_by_type[input_type](*importer_params)
    result = asset_importer.import_assets(assets=assets)  # type:ignore X
    return result



"""Service for managing plugins in kili"""

from kili.services.plugins.activation import activate_plugin, deactivate_plugin
from kili.services.plugins.deletion import delete_plugin
from kili.services.plugins.upload import PluginUploader, WebhookUploader

__all__ = [
    "PluginUploader",
    "WebhookUploader",
    "activate_plugin",
    "deactivate_plugin",
    "delete_plugin",
]



"""
Gets the
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List

import requests

from .exceptions import DownloadError


class AbstractContentRepository(ABC):
    """
    Interface to the content repository
    """

    def __init__(
        self, router_endpoint: str, router_headers: Dict[str, str], verify_ssl: bool
    ) -> None:
        self.router_endpoint = router_endpoint
        self.router_headers = router_headers
        self.verify_ssl = verify_ssl
        assert router_endpoint, "The router endpoint string should not be empty"

    @abstractmethod
    def get_frames(self, content_url: str) -> List[str]:
        """
        Get asset content frames.
        """

    @abstractmethod
    def get_content_stream(self, content_url: str, block_size: int) -> Iterator[Any]:
        """
        Get asset content stream.
        """

    def get_content_frames_paths(self, asset: Dict) -> List[str]:
        """
        Get list of links to frames from the file located at asset[jsonContent].
        Returns an empty list if `content` in the asset exists.
        """
        content_frames = []

        if not asset["content"] and asset["jsonContent"]:
            content_frames = self.get_frames(asset["jsonContent"])

        return content_frames

    def is_serving(self, url: str) -> bool:
        """
        Return a boolean defining if the asset is served by Kili or not
        """
        return url.startswith(self.router_endpoint)


class SDKContentRepository(AbstractContentRepository):
    """
    Handle content fetching from the server from the SDK
    """

    def get_frames(self, content_url: str) -> List[str]:
        frames: List[str] = []
        headers = None
        if content_url.startswith(self.router_endpoint):
            headers = self.router_headers
        json_content_resp = requests.get(
            content_url, headers=headers, verify=self.verify_ssl, timeout=30
        )

        if json_content_resp.ok:
            frames = list(json_content_resp.json().values())
        return frames

    def get_content_stream(self, content_url: str, block_size: int) -> Iterator[Any]:
        response = requests.get(
            content_url,
            stream=True,
            headers=None,
            verify=self.verify_ssl,
            timeout=30,
        )

        if not response.ok:
            response = requests.get(
                content_url,
                stream=True,
                headers=self.router_headers,  # pass the API key if first request failed
                verify=self.verify_ssl,
                timeout=30,
            )

            if not response.ok:
                raise DownloadError(f"Error while downloading image {content_url}")

        return response.iter_content(block_size)



"""
GraphQL Queries of Labels
"""


from typing import List, Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery
from kili.orm import Label as LabelFormatType


class LabelWhere(BaseQueryWhere):
    """
    Tuple to be passed to the LabelQuery to restrict query
    """

    # pylint: disable=too-many-arguments,too-many-locals,too-many-instance-attributes

    def __init__(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        category_search: Optional[str] = None,
    ):
        self.project_id = project_id
        self.asset_id = asset_id
        self.asset_status_in = asset_status_in
        self.asset_external_id_in = asset_external_id_in
        self.author_in = author_in
        self.created_at = created_at
        self.created_at_gte = created_at_gte
        self.created_at_lte = created_at_lte
        self.honeypot_mark_gte = honeypot_mark_gte
        self.honeypot_mark_lte = honeypot_mark_lte
        self.id_contains = id_contains
        self.label_id = label_id
        self.type_in = type_in
        self.user_id = user_id
        self.category_search = category_search
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK LabelWhere"""
        return {
            "id": self.label_id,
            "asset": {
                "id": self.asset_id,
                "externalIdStrictlyIn": self.asset_external_id_in,
                "statusIn": self.asset_status_in,
            },
            "project": {
                "id": self.project_id,
            },
            "user": {
                "id": self.user_id,
            },
            "createdAt": self.created_at,
            "createdAtGte": self.created_at_gte,
            "createdAtLte": self.created_at_lte,
            "authorIn": self.author_in,
            "honeypotMarkGte": self.honeypot_mark_gte,
            "honeypotMarkLte": self.honeypot_mark_lte,
            "idIn": self.id_contains,
            "search": self.category_search,
            "typeIn": self.type_in,
        }


class LabelQuery(GraphQLQuery):
    """Label query."""

    FORMAT_TYPE = LabelFormatType

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL labels query
        """
        return f"""
        query labels($where: LabelWhere!, $first: PageSize!, $skip: Int!) {{
            data: labels(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countLabels($where: LabelWhere!) {
        data: countLabels(where: $where)
    }
    """



"""
Helpers for the asset mutations
"""
from typing import Dict, List, Optional

from kili.authentication import KiliAuth
from kili.mutations.helpers import check_asset_identifier_arguments
from kili.services.helpers import infer_ids_from_external_ids

from ...helpers import convert_to_list_of_none, format_metadata, is_none_or_empty


def process_update_properties_in_assets_parameters(properties) -> Dict:
    """
    Process arguments of the update_properties_in_assets method
    and return the properties for the paginated loop
    """
    formatted_json_metadatas = None
    if properties["json_metadatas"] is None:
        formatted_json_metadatas = None
    else:
        if isinstance(properties["json_metadatas"], list):
            formatted_json_metadatas = list(map(format_metadata, properties["json_metadatas"]))
        else:
            raise TypeError(
                "json_metadatas",
                "Should be either a None or a list of None, string, list or dict",
            )
    properties["json_metadatas"] = formatted_json_metadatas
    nb_assets_to_modify = len(properties["asset_ids"])
    properties = {
        k: convert_to_list_of_none(v, length=nb_assets_to_modify) for k, v in properties.items()
    }
    properties["should_reset_to_be_labeled_by_array"] = list(
        map(is_none_or_empty, properties["to_be_labeled_by_array"])
    )
    return properties


def get_asset_ids_or_throw_error(
    auth: KiliAuth,
    asset_ids: Optional[List[str]],
    external_ids: Optional[List[str]],
    project_id: Optional[str],
) -> List[str]:
    """
    Check if external id to internal id conversion is valid and needed.
    """
    check_asset_identifier_arguments(project_id, asset_ids, external_ids)

    if asset_ids is None:
        id_map = infer_ids_from_external_ids(
            auth=auth, asset_external_ids=external_ids, project_id=project_id  # type: ignore
        )
        asset_ids = [id_map[id] for id in external_ids]  # type: ignore

    return asset_ids



"""Project mutations."""
from typing import List, Optional

from typeguard import typechecked

from kili.authentication import KiliAuth
from kili.services.plugins import (
    PluginUploader,
    WebhookUploader,
    activate_plugin,
    deactivate_plugin,
    delete_plugin,
)
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class MutationsPlugins:
    """Set of Plugins mutations."""

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def upload_plugin(
        self,
        plugin_path: Optional[str] = None,
        plugin_name: Optional[str] = None,
        verbose: bool = True,
        **kwargs
    ):
        """Uploads a plugin.

        Args:
            plugin_path : Path to your plugin. Either:

                - a folder containing a main.py (mandatory) and a requirements.txt (optional)
                - a .py file
            plugin_name: name of your plugin, if not provided, it will be the name from your file
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.upload_plugin(plugin_path="./path/to/my/folder")
            >>> kili.upload_plugin(plugin_path="./path/to/my/file.py")
        """

        if kwargs.get("file_path"):
            raise TypeError(
                '"file_path" has been deprecated for "plugin_path", please use "plugin_path"'
                " instead"
            )

        if not plugin_path:
            raise TypeError('"plugin_path is nullish, please provide a value')

        return PluginUploader(self.auth, plugin_path, plugin_name, verbose).create_plugin()

    @typechecked
    def create_webhook(
        self,
        webhook_url: str,
        plugin_name: str,
        header: Optional[str] = None,
        verbose: bool = True,
        handler_types: Optional[List[str]] = None,
    ):
        # pylint: disable=line-too-long,too-many-arguments
        """
        Create a webhook linked to Kili's events.
        For a complete example, refer to the notebook `webhooks_example` on kili repo

        Args:
            webhook_url: URL receiving post requests on events on Kili. The payload will be the following:

                - eventType: the type of event called
                - logPayload:
                    - runId: a unique identifier of the run for observability
                    - projectId: the Kili project the webhook is called on
                - payload: the event produced, for example for `onSubmit` event:
                    - label: the label produced
                    - asset_id: the asset on which the label is produced
            plugin_name: name of your plugin
            header: Authorization header to access the routes
            verbose: If false, minimal logs are displayed
            handler_types: List of actions for which the webhook should be called.
                Possible variants: `onSubmit`, `onReview`.
                By default, is [`onSubmit`, `onReview`].

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.create_webhook(webhook_url='https://my-custom-url-publicly-accessible/', plugin_name='my webhook', header='...')
        """

        return WebhookUploader(
            self.auth, webhook_url, plugin_name, header, verbose, handler_types
        ).create_webhook()

    @typechecked
    def update_webhook(
        self,
        new_webhook_url: str,
        plugin_name: str,
        new_header: Optional[str] = None,
        verbose: bool = True,
        handler_types: Optional[List[str]] = None,
    ):
        # pylint: disable=line-too-long,too-many-arguments
        """
        Update a webhook linked to Kili's events.
        For a complete example, refer to the notebook `webhooks_example` on kili repo

        Args:
            new_webhook_url: New URL receiving post requests on events on Kili. See `create_webhook` for the payload description
            plugin_name: name of your plugin
            new_header: Authorization header to access the routes
            verbose: If false, minimal logs are displayed
            handler_types: List of actions for which the webhook should be called.
                Possible variants: `onSubmit`, `onReview`.
                By default, is [`onSubmit`, `onReview`]

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_webhook(webhook_url='https://my-custom-url-publicly-accessible/', plugin_name='my webhook', header='...')
        """

        return WebhookUploader(
            self.auth,
            new_webhook_url,
            plugin_name,
            new_header,
            verbose,
            handler_types,
        ).update_webhook()

    @typechecked
    def activate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.activate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return activate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def deactivate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.deactivate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return deactivate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def delete_plugin(
        self,
        plugin_name: str,
    ):
        """Deletes a plugin.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.delete_plugin(plugin_name="my_plugin_name")
        """

        return delete_plugin(self.auth, plugin_name)

    @typechecked
    def update_plugin(
        self,
        plugin_path: Optional[str] = None,
        plugin_name: Optional[str] = None,
        verbose: bool = True,
        **kwargs
    ):
        """Update a plugin with new code.

        Args:
            plugin_path: Path to your plugin. Either:

                - a folder containing a main.py (mandatory) and a requirements.txt (optional)
                - a .py file
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_plugin(plugin_name="my_plugin_name")
        """

        if kwargs.get("file_path"):
            raise TypeError(
                """ "file_path" has been deprecated for "plugin_path",
                please use "plugin_path" instead"""
            )

        if not plugin_path:
            raise TypeError('"plugin_path is nullish, please provide a value')

        if not plugin_name:
            raise TypeError('"plugin_name is nullish, please provide a value')

        return PluginUploader(self.auth, plugin_path, plugin_name, verbose).update_plugin()



"""
GraphQL Queries of data integrations
"""


from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery


class DataIntegrationWhere(BaseQueryWhere):
    """
    Tuple to be passed to the DataIntegrationsQuery to restrict the query
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        data_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[str] = None,
        status: Optional[str] = None,
        organization_id: Optional[str] = None,
    ) -> None:
        self.data_integration_id = data_integration_id
        self.name = name
        self.platform = platform
        self.status = status
        self.organization_id = organization_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK DataIntegrationWhere"""
        return {
            "id": self.data_integration_id,
            "name": self.name,
            "platform": self.platform,
            "status": self.status,
            "organizationId": self.organization_id,
        }


class DataIntegrationsQuery(GraphQLQuery):
    """DataIntegrations query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL dataIntegrations query
        """
        return f"""
        query dataIntegrations($where: DataIntegrationWhere!, $first: PageSize!, $skip: Int!) {{
          data: dataIntegrations(where: $where, skip: $skip, first: $first) {{
            {fragment}
          }}
        }}
        """

    COUNT_QUERY = """
    query countDataIntegrations($where: DataIntegrationWhere!) {
      data: countDataIntegrations(where: $where)
    }
    """



"""Common arguments and options for the CLI"""
from typing import List, Union

import click

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
ROLES = ["ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"]


class BoolOrStringParamType(click.ParamType):
    """Click parameter type for boolean or string. This is useful for the --ssl-verify option,
    that can take a boolean or a string as value (like the `requests` library)"""

    name = "bool_or_string"

    def convert(self, value, param, ctx) -> Union[bool, str]:
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False
        return value


BOOL_OR_STRING = BoolOrStringParamType()


class Options:  # pylint: disable=too-few-public-methods
    """Common options for the CLI"""

    api_key = click.option(
        "--api-key",
        type=str,
        default=None,
        help="Your Kili API key. ",
        show_default='"KILI_API_KEY" environment variable',
    )

    endpoint = click.option(
        "--endpoint",
        type=str,
        default=None,
        help="Kili API Endpoint. ",
        show_default=(
            '"KILI_API_ENDPOINT" environment variable,'
            ' Kili SAAS: "https://cloud.kili-technology.com/api/label/v2/graphql"'
        ),
    )

    project_id = click.option("--project-id", type=str, required=True, help="Id of the project")

    tablefmt = click.option(
        "--stdout-format",
        "tablefmt",
        type=str,
        default="plain",
        help=(
            "Defines how the output table is formatted "
            "(see https://pypi.org/project/tabulate/, default: plain)."
        ),
    )

    from_project = click.option(
        "--from-project",
        "project_id_src",
        type=str,
        help="project_id of another Kili project",
    )

    role = click.option(
        "--role",
        type=click.Choice(ROLES),
        default=None,
        show_default="LABELER",
        help="Project role of the added user(s).",
    )

    script_name = click.option(
        "--script-name",
        type=str,
        default=None,
        help=(
            "Script's name in Kili database."
            "If not provided, it will be the file name on your machine"
        ),
    )

    verbose = click.option(
        "--verbose", type=bool, is_flag=True, default=False, help="Show more logs"
    )

    ssl_verify = click.option(
        "--ssl-verify",
        type=BOOL_OR_STRING,
        default="True",
        help=(
            "Verify SSL certificate (see Python's requests' verify argument documentation for more"
            " details:"
            " https://requests.readthedocs.io/en/latest/user/advanced/#ssl-cert-verification)"
        ),
    )


def from_csv(required_columns: List[str], optionnal_columns: List[str]):
    """--from-csv shared click option"""
    help_ = (
        "path to a csv file with required columns:"
        + ", ".join(required_columns)
        + " required columns: "
        + ", ".join(optionnal_columns)
    )

    return click.option(
        "--from-csv",
        "csv_path",
        type=click.Path(),
        help=help_,
    )


class Arguments:  # pylint: disable=too-few-public-methods
    """Common arguments for the CLI"""

    files = click.argument("files", type=click.Path(), required=False, nargs=-1)

    emails = click.argument("emails", type=str, required=False, nargs=-1)

    project_id = click.argument("project_id", type=str, required=True)



"""Label subscription."""

from dataclasses import dataclass
from typing import Callable

from typeguard import typechecked

from kili.authentication import KiliAuth

from ...graphql.graphql_client import SubscriptionGraphQLClient
from .subscriptions import GQL_LABEL_CREATED_OR_UPDATED


@dataclass
class SubscriptionsLabel:
    """Set of Label subscriptions."""

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def label_created_or_updated(
        self, project_id: str, callback: Callable[[str, str], None]
    ) -> SubscriptionGraphQLClient:
        # pylint: disable=line-too-long
        """
        Subscribe a callback to a project, which is executed when a label is created or updated.

        Args:
            project_id: Identifier of the project
            callback: This function takes as input the id of the asset and its content.

        Returns:
            A subscription client.

        !!! example "Recipe"
            For more detailed examples on how to use Webhooks,
            See [the related recipe](https://github.com/kili-technology/kili-python-sdk/blob/master/recipes/webhooks.ipynb)
        """
        ws_endpoint = self.auth.client.endpoint.replace("http", "ws")
        websocket = SubscriptionGraphQLClient(ws_endpoint)
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        authorization = f"X-API-Key: {self.auth.api_key}"
        headers["Authorization"] = authorization
        variables = {"projectID": project_id}
        websocket.subscribe(
            GQL_LABEL_CREATED_OR_UPDATED,
            variables=variables,
            callback=callback,
            headers=headers,
            authorization=authorization,
        )
        return websocket



"""
GraphQL module
"""

from abc import ABC, abstractmethod
from typing import Callable, Dict, Generator, List, NamedTuple, Optional, Type, Union

from tqdm import tqdm
from typeguard import typechecked

from kili.constants import QUERY_BATCH_SIZE
from kili.helpers import format_result
from kili.utils.pagination import api_throttle

from .graphql_client import GraphQLClient


class QueryOptions(NamedTuple):
    """Options when calling GraphQLQuery from the SDK"""

    disable_tqdm: Optional[bool]
    first: Optional[int] = None
    skip: int = 0


class BaseQueryWhere(ABC):
    """
    Abtsract class for defining the where payload to send in a graphQL query
    """

    def __init__(self):
        self._graphql_payload = self.graphql_where_builder()

    @abstractmethod
    def graphql_where_builder(self) -> Dict:
        """Build the GraphQL where payload sent in the resolver from the
        arguments given to the where class
        """
        raise NotImplementedError

    @property
    def graphql_payload(self):
        """where payload to send in the graphQL query"""
        return self._graphql_payload


class GraphQLQuery(ABC):
    """
    Query class for querying Kili objects.
    It factorizes code for executing paginated queries
    """

    def __init__(
        self,
        client: GraphQLClient,
        ssl_verify: Union[bool, str],
    ) -> None:
        self.client = client
        self.ssl_verify = ssl_verify

    @staticmethod
    @abstractmethod
    def query(fragment: str) -> str:
        """Return the GraphQL object query to be sent, according to the fragment to return.

        Args:
            fragment: the fragment to return in the query

        Return:
            The query to be sent to graphQL
        """
        raise NotImplementedError

    COUNT_QUERY: str = NotImplemented
    FORMAT_TYPE: Optional[Type] = None

    def __call__(
        self,
        where: BaseQueryWhere,
        fields: List[str],
        options: QueryOptions,
        post_call_function: Optional[Callable] = None,
    ) -> Generator[Dict, None, None]:
        """Get a generator of objects of the specified type in accordance with the provided where"""
        fragment = self.fragment_builder(fields)
        query = self.query(fragment)

        return self.execute_query_from_paginated_call(query, where, options, post_call_function)

    def count(self, where: BaseQueryWhere) -> int:
        """Count the number of objects matching the given where payload"""
        payload = {"where": where.graphql_payload}
        count_result = self.client.execute(self.COUNT_QUERY, payload)
        return format_result("data", count_result, int, self.ssl_verify)

    def get_number_of_elements_to_query(self, where: BaseQueryWhere, options: QueryOptions):
        """Return the total number of element to query for one query.
        It uses both the argument first given by the user
        and the total number of available objects obtained with a graphQL count query
        """
        first = options.first
        skip = options.skip
        count_rows_available = self.count(where)
        count_objects_queried = max(count_rows_available - skip, 0)
        if first is None:
            return count_objects_queried
        return min(count_objects_queried, first)

    def execute_query_from_paginated_call(
        self,
        query: str,
        where: BaseQueryWhere,
        options: QueryOptions,
        post_call_function: Optional[Callable],
    ) -> Generator[Dict, None, None]:
        """
        Builds a row generator from paginated calls.

        Args:
            query: The object query to execute and to send to graphQL, in string format
            where_payload: The where payload to be sent to graphQL,
                as a value of the 'where' key in the global payload
            options: The query options
        """
        # we can get the total number of elements to query
        if isinstance(self.COUNT_QUERY, str):
            nb_rows_to_query = self.get_number_of_elements_to_query(where, options)
            disable_tqdm = options.disable_tqdm

        # we don't have count methods but we know the total number of elements to query
        elif options.first is not None:
            nb_rows_to_query = options.first
            disable_tqdm = options.disable_tqdm

        # we don't have count methods and we don't know the total number of elements to query
        else:
            nb_rows_to_query = None
            disable_tqdm = True

        if nb_rows_to_query == 0:
            yield from ()
        else:
            with tqdm(total=nb_rows_to_query, disable=disable_tqdm) as pbar:
                count_rows_retrieved = 0
                while True:
                    if nb_rows_to_query is not None and count_rows_retrieved >= nb_rows_to_query:
                        break

                    skip = count_rows_retrieved + options.skip
                    first = (
                        min(QUERY_BATCH_SIZE, nb_rows_to_query - count_rows_retrieved)
                        if nb_rows_to_query is not None
                        else QUERY_BATCH_SIZE
                    )
                    payload = {"where": where.graphql_payload, "skip": skip, "first": first}
                    rows = api_throttle(self.client.execute)(query, payload)
                    rows = format_result("data", rows, self.FORMAT_TYPE, self.ssl_verify)

                    if rows is None or len(rows) == 0:
                        break

                    if post_call_function is not None:
                        rows = post_call_function(rows)

                    if isinstance(rows, Dict):
                        yield rows
                        break

                    for row in rows:
                        yield row

                    count_rows_retrieved += len(rows)
                    pbar.update(len(rows))

                    if len(rows) < first:
                        break

    @typechecked
    def fragment_builder(self, fields: List[str]):
        """
        Builds a GraphQL fragment for a list of fields to query

        Args:
            fields: The list of fields to query
        """
        fragment = ""

        # split a field and its subfields (e.g. "roles.user.id" -> ["roles", "user.id"])
        subfields = [field.split(".", 1) for field in fields if "." in field]

        if subfields:
            # get the root fields (e.g. "roles" in "roles.user.id")
            root_fields = {subfield[0] for subfield in subfields}
            for root_field in root_fields:
                # get the subfields of the root field (e.g. "user.id" in "roles.user.id")
                fields_subquery = [
                    subfield[1] for subfield in subfields if subfield[0] == root_field
                ]
                # build the subquery fragment (e.g. "user{id}" in "roles{user{id}}")
                new_fragment = self.fragment_builder(fields_subquery)
                # add the subquery to the fragment
                fragment += f" {root_field}{{{new_fragment}}}"

            # remove the fields that have been queried in subqueries (e.g. "roles.user.id")
            fields = [field for field in fields if "." not in field]

        for field in fields:
            fragment += f" {field}"

        return fragment



"""Data connection mutations."""

from datetime import datetime
from typing import Dict, List, Optional

from typeguard import typechecked

from kili.authentication import KiliAuth
from kili.exceptions import GraphQLError
from kili.utils.logcontext import for_all_methods, log_call

from ... import services
from ...helpers import format_result
from ...queries.data_integration.queries import (
    GQL_GET_DATA_INTEGRATION_FOLDER_AND_SUBFOLDERS,
)
from .exceptions import AddDataConnectionError
from .queries import GQL_ADD_PROJECT_DATA_CONNECTION


@for_all_methods(log_call, exclude=["__init__"])
class MutationsDataConnection:
    """Set of DataConnection mutations."""

    def __init__(self, auth: KiliAuth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def add_cloud_storage_connection(
        self,
        project_id: str,
        cloud_storage_integration_id: str,
        selected_folders: Optional[List[str]] = None,
    ) -> Dict:
        """Connect a cloud storage to a project.

        Args:
            project_id: ID of the project.
            cloud_storage_integration_id: ID of the cloud storage integration.
            selected_folders: List of folders of the data integration to connect to the project.
                If not provided, all folders of the data integration will be connected.

        Returns:
            A dict with the DataConnection ID.
        """
        if selected_folders is None:
            variables = {"dataIntegrationId": cloud_storage_integration_id}
            try:
                result = self.auth.client.execute(
                    GQL_GET_DATA_INTEGRATION_FOLDER_AND_SUBFOLDERS, variables=variables
                )
            except GraphQLError as err:
                raise AddDataConnectionError(
                    f"The data integration with id {cloud_storage_integration_id} is not supported"
                    " in the SDK yet. Use the Kili app to create a data connection instead."
                ) from err
            result = format_result("data", result, None, self.auth.ssl_verify)
            selected_folders = [folder["key"] for folder in result]

        variables = {
            "data": {
                "projectId": project_id,
                "integrationId": cloud_storage_integration_id,
                "isChecking": False,
                "lastChecked": datetime.now().isoformat(sep="T", timespec="milliseconds") + "Z",
                "selectedFolders": selected_folders,
            }
        }
        result = self.auth.client.execute(GQL_ADD_PROJECT_DATA_CONNECTION, variables)
        result = format_result("data", result, None, self.auth.ssl_verify)

        # We trigger data difference computation (same behavior as in the frontend)
        services.compute_differences(self.auth, result["id"])

        return result

    @typechecked
    def synchronize_cloud_storage_connection(
        self,
        cloud_storage_connection_id: str,
        delete_extraneous_files: bool = False,
    ) -> Dict:
        """Synchronize a cloud storage connection.

        This method will compute differences between the cloud storage connection and the project
            and then validate the differences.

        If `delete_extraneous_files` is True, it will also delete files that are not in the
            cloud storage integration anymore but that are still in the project.

        Args:
            cloud_storage_connection_id: ID of the cloud storage connection.
            delete_extraneous_files: If True, delete extraneous files.

        Returns:
            A dict with the DataConnection ID.
        """
        return services.synchronize_data_connection(
            self.auth, cloud_storage_connection_id, delete_extraneous_files
        )



"""
Types used by the conversion service
"""
from typing import Any, Dict, List, NamedTuple, NewType, Tuple, Union

from typing_extensions import Literal, TypedDict

InputType = Literal["TEXT", "IMAGE"]
AssetId = NewType("AssetId", str)
ProjectId = NewType("ProjectId", str)

LogLevel = Union[int, Literal["ERROR", "WARNING", "DEBUG", "INFO", "CRITICAL"]]


class JobCategory(NamedTuple):
    """
    Contains information for a category
    """

    category_name: str
    id: int
    job_id: str


YoloAnnotation = Tuple[int, float, float, float, float]
LabelType = Literal["AUTOSAVE", "DEFAULT", "PREDICTION", "REVIEW", "INFERENCE"]

MLTask = Literal["CLASSIFICATION", "NAMED_ENTITIES_RECOGNITION", "OBJECT_DETECTION"]
Tool = Literal["rectangle", "semantic", "polygon"]
JobName = NewType("JobName", str)


class Job(TypedDict):
    """
    Contains job settings
    """

    content: Any
    instruction: str
    isChild: bool
    tools: List[Tool]
    mlTask: MLTask
    models: Any  # example: {"interactive-segmentation": {"job": "SEMANTIC_JOB_MARKER"}},
    isVisible: bool
    required: int
    isNew: bool


Jobs = Dict[JobName, Job]



"""
Label mutations
"""

import warnings
from json import dumps
from typing import Dict, List, Optional

from typeguard import typechecked

from kili import services
from kili.authentication import KiliAuth
from kili.helpers import deprecate, format_result
from kili.mutations.helpers import check_asset_identifier_arguments
from kili.mutations.label.queries import (
    GQL_APPEND_TO_LABELS,
    GQL_CREATE_HONEYPOT,
    GQL_UPDATE_PROPERTIES_IN_LABEL,
)
from kili.orm import Label
from kili.services.helpers import (
    assert_all_arrays_have_same_size,
    infer_ids_from_external_ids,
)
from kili.services.types import LabelType
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class MutationsLabel:
    """Set of Label mutations."""

    # pylint: disable=too-many-arguments

    def __init__(self, auth: KiliAuth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def create_predictions(
        self,
        project_id: str,
        external_id_array: Optional[List[str]] = None,
        model_name_array: Optional[List[str]] = None,
        json_response_array: Optional[List[dict]] = None,
        model_name: Optional[str] = None,
        asset_id_array: Optional[List[str]] = None,
    ) -> Dict:
        # pylint: disable=line-too-long
        """Create predictions for specific assets.

        Args:
            project_id: Identifier of the project.
            external_id_array: The external IDs of the assets for which we want to add predictions.
            model_name_array: Deprecated, use `model_name` instead.
            json_response_array: The predictions are given here. For examples,
                see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).
            model_name: The name of the model that generated the predictions
            asset_id_array: The internal IDs of the assets for which we want to add predictions.

        Returns:
            A result object which indicates if the mutation was successful, or an error message.

        !!! example "Recipe"
            For more detailed examples on how to create predictions, see [the recipe](https://docs.kili-technology.com/recipes/importing-labels-and-predictions).

        !!! warning "model name"
            The use of `model_name_array` is deprecated. Creating predictions from different
            models is not supported anymore. Please use `model_name` argument instead to
            provide the predictions model name.
        """
        if json_response_array is None or len(json_response_array) == 0:
            raise ValueError(
                "json_response_array is empty, you must provide at least one prediction to upload"
            )
        assert_all_arrays_have_same_size(
            [external_id_array, json_response_array, model_name_array, asset_id_array]
        )
        if model_name is None:
            if model_name_array is None:
                raise ValueError("You must provide a model name with the model_name argument ")
            if len(set(model_name_array)) > 1:
                raise ValueError(
                    "Creating predictions from different models is not supported anymore. Separate"
                    " your calls by models."
                )
            warnings.warn(
                (
                    "The use of `model_name_array` is deprecated. Creating predictions from"
                    " different models is not supported anymore. Please use `model_name` argument"
                    " instead to provide the predictions model name."
                ),
                DeprecationWarning,
                stacklevel=1,
            )
            model_name = model_name_array[0]

        labels = [
            {
                "asset_id": asset_id,
                "asset_external_id": asset_external_id,
                "json_response": json_response,
            }
            for (asset_id, asset_external_id, json_response) in list(
                zip(
                    asset_id_array or [None] * len(json_response_array),
                    external_id_array or [None] * len(json_response_array),
                    json_response_array,
                )
            )
        ]
        services.import_labels_from_dict(self.auth, project_id, labels, "PREDICTION", model_name)
        return {"id": project_id}

    @deprecate(
        msg=(
            "append_to_labels method is deprecated. Please use append_labels instead. This new"
            " function allows to import several labels 10 times faster."
        )
    )
    @typechecked
    def append_to_labels(
        self,
        json_response: dict,
        author_id: Optional[str] = None,
        label_asset_external_id: Optional[str] = None,
        label_asset_id: Optional[str] = None,
        label_type: LabelType = "DEFAULT",
        project_id: Optional[str] = None,
        seconds_to_label: Optional[int] = 0,
    ):
        """
        !!! danger "[DEPRECATED]"
            append_to_labels method is deprecated. Please use append_labels instead.
            This new function allows to import several labels 10 times faster.

        Append a label to an asset.

        Args:
            json_response: Label is given here
            author_id: ID of the author of the label
            label_asset_external_id: External identifier of the asset
            label_asset_id: Identifier of the asset
            project_id: Identifier of the project
            label_type: Can be one of `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`
            seconds_to_label: Time to create the label

        !!! warning
            Either provide `label_asset_id` or `label_asset_external_id` and `project_id`

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(label_asset_id=asset_id, json_response={...})

        """
        if author_id is None:
            author_id = self.auth.user_id
        check_asset_identifier_arguments(
            project_id,
            [label_asset_id] if label_asset_id else None,
            [label_asset_external_id] if label_asset_external_id else None,
        )
        if label_asset_id is None:
            assert label_asset_external_id and project_id
            label_asset_id = infer_ids_from_external_ids(
                self.auth, [label_asset_external_id], project_id
            )[label_asset_external_id]
        variables = {
            "data": {
                "authorID": author_id,
                "jsonResponse": dumps(json_response),
                "labelType": label_type,
                "secondsToLabel": seconds_to_label,
            },
            "where": {"id": label_asset_id},
        }
        result = self.auth.client.execute(GQL_APPEND_TO_LABELS, variables)
        return format_result("data", result, Label, self.auth.ssl_verify)

    @typechecked
    def append_labels(  # pylint: disable=dangerous-default-value
        self,
        asset_id_array: Optional[List[str]] = None,
        json_response_array: List[Dict] = [],
        author_id_array: Optional[List[str]] = None,
        seconds_to_label_array: Optional[List[int]] = None,
        model_name: Optional[str] = None,
        label_type: LabelType = "DEFAULT",
        project_id: Optional[str] = None,
        asset_external_id_array: Optional[List[str]] = None,
        disable_tqdm: bool = False,
    ) -> List:
        """Append labels to assets.

        Args:
            asset_id_array: list of asset internal ids to append labels on
            json_response_array: list of labels to append
            author_id_array: list of the author id of the labels
            seconds_to_label_array: list of times taken to produce the label, in seconds
            model_name: Only useful when uploading predictions.
                Name of the model when uploading predictions
            label_type: Can be one of `AUTOSAVE`, `DEFAULT`, `PREDICTION`, `REVIEW` or `INFERENCE`
            project_id: Identifier of the project
            asset_external_id_array: list of asset external ids to append labels on

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.append_to_labels(
                    asset_id_array=['cl9wmlkuc00050qsz6ut39g8h', 'cl9wmlkuw00080qsz2kqh8aiy'],
                    json_response_array=[{...}, {...}]
                )
        """
        if len(json_response_array) == 0:
            raise ValueError(
                "json_response_array is empty, you must provide at least one label to upload"
            )
        check_asset_identifier_arguments(project_id, asset_id_array, asset_external_id_array)
        assert_all_arrays_have_same_size(
            [
                seconds_to_label_array,
                author_id_array,
                json_response_array,
                asset_external_id_array,
                asset_id_array,
            ]
        )

        labels = [
            {
                "asset_id": asset_id,
                "asset_external_id": asset_external_id,
                "json_response": json_response,
                "seconds_to_label": seconds_to_label,
                "author_id": author_id,
            }
            for (asset_id, asset_external_id, json_response, seconds_to_label, author_id) in list(
                zip(
                    asset_id_array or [None] * len(json_response_array),
                    asset_external_id_array or [None] * len(json_response_array),
                    json_response_array,
                    seconds_to_label_array or [None] * len(json_response_array),
                    author_id_array or [None] * len(json_response_array),
                )
            )
        ]
        return services.import_labels_from_dict(
            self.auth, project_id, labels, label_type, model_name, disable_tqdm
        )

    @typechecked
    def update_properties_in_label(
        self,
        label_id: str,
        seconds_to_label: Optional[int] = None,
        model_name: Optional[str] = None,
        json_response: Optional[dict] = None,
    ) -> Label:
        """Update properties of a label.

        Args:
            label_id: Identifier of the label
            seconds_to_label: Time to create the label
            model_name: Name of the model
            json_response: The label is given here

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_properties_in_label(label_id=label_id, json_response={...})
        """
        formatted_json_response = None if json_response is None else dumps(json_response)
        variables = {
            "labelID": label_id,
            "secondsToLabel": seconds_to_label,
            "modelName": model_name,
            "jsonResponse": formatted_json_response,
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_LABEL, variables)
        return format_result("data", result, Label, self.auth.ssl_verify)

    @typechecked
    def create_honeypot(
        self,
        json_response: dict,
        asset_external_id: Optional[str] = None,
        asset_id: Optional[str] = None,
        project_id: Optional[str] = None,
    ) -> Label:
        """Create honeypot for an asset.

        !!! info
            Uses the given `json_response` to create a `REVIEW` label.
            This enables Kili to compute a`honeypotMark`,
            which measures the similarity between this label and other labels.

        Args:
            json_response: The JSON response of the honeypot label of the asset
            asset_id: Identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            asset_external_id: External identifier of the asset
                Either provide asset_id or asset_external_id and project_id
            project_id: Identifier of the project
                Either provide asset_id or asset_external_id and project_id

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        if asset_id is None:
            if asset_external_id is None or project_id is None:
                raise ValueError("Either provide asset_id or external_id and project_id")
            asset_id = infer_ids_from_external_ids(self.auth, [asset_external_id], project_id)[
                asset_external_id
            ]

        variables = {
            "data": {"jsonResponse": dumps(json_response)},
            "where": {"id": asset_id},
        }
        result = self.auth.client.execute(GQL_CREATE_HONEYPOT, variables)
        return format_result("data", result, Label, self.auth.ssl_verify)



"""
Base class for all formatters and other utility classes.
"""

import csv
import json
import logging
import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple, cast

from kili.authentication import KiliAuth
from kili.orm import Asset, Label
from kili.services.export.repository import AbstractContentRepository
from kili.services.export.tools import fetch_assets
from kili.services.export.types import (
    CocoAnnotationModifier,
    ExportType,
    LabelFormat,
    SplitOption,
)
from kili.services.project import get_project
from kili.services.types import Job, ProjectId
from kili.utils.tempfile import TemporaryDirectory


class ExportParams(NamedTuple):
    """
    Contains all parameters that change the result of the export
    """

    assets_ids: Optional[List[str]]
    export_type: ExportType
    project_id: ProjectId
    label_format: LabelFormat
    split_option: SplitOption
    single_file: bool
    output_file: Path
    with_assets: bool
    annotation_modifier: Optional[CocoAnnotationModifier]


class AbstractExporter(ABC):  # pylint: disable=too-many-instance-attributes
    """
    Abstract class defining the interface for all exporters.
    """

    def __init__(
        self,
        export_params: ExportParams,
        auth: KiliAuth,
        logger: logging.Logger,
        disable_tqdm: bool,
        content_repository: AbstractContentRepository,
    ):  # pylint: disable=too-many-arguments
        self.project_id: ProjectId = export_params.project_id
        self.assets_ids: Optional[List[str]] = export_params.assets_ids
        self.export_type: ExportType = export_params.export_type
        self.label_format: LabelFormat = export_params.label_format
        self.single_file: bool = export_params.single_file
        self.split_option: SplitOption = export_params.split_option
        self.disable_tqdm: bool = disable_tqdm
        self.auth = auth
        self.logger: logging.Logger = logger
        self.content_repository: AbstractContentRepository = content_repository
        self.output_file = export_params.output_file
        self.with_assets: bool = export_params.with_assets
        self.export_root_folder: Path = Path()
        self.annotation_modifier = export_params.annotation_modifier

        project_info = get_project(
            self.auth, self.project_id, ["jsonInterface", "inputType", "title"]
        )
        self.project_json_interface = project_info["jsonInterface"]
        self.project_input_type = project_info["inputType"]
        self.project_title = project_info["title"]

    @abstractmethod
    def _check_arguments_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the export options.
        """

    @abstractmethod
    def _check_project_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the project.
        """

    def _is_job_compatible(self, job: Job) -> bool:
        """
        Check if the export label format is compatible with the job.
        """
        raise NotImplementedError

    @property
    def compatible_jobs(self) -> Tuple[str]:
        """
        Get all job names compatible with the export format.
        """
        return tuple(
            job_name
            for job_name, job in self.project_json_interface["jobs"].items()
            if self._is_job_compatible(job)
        )

    @abstractmethod
    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """
        Converts the asset and save them into an archive file.
        """

    def make_archive(self, root_folder: Path, output_filename: Path) -> Path:
        """
        Make the export archive
        """
        path_folder = root_folder / self.project_id
        path_archive = shutil.make_archive(str(path_folder), "zip", path_folder)
        output_filename.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(path_archive, output_filename)
        return output_filename

    def create_readme_kili_file(self, root_folder: Path) -> None:
        """
        Create a README.kili.txt file to give information about exported labels
        """
        readme_file_name = root_folder / self.project_id / "README.kili.txt"
        project_info = get_project(self.auth, self.project_id, ["title", "id", "description"])
        readme_file_name.parent.mkdir(parents=True, exist_ok=True)
        with readme_file_name.open("wb") as fout:
            fout.write("Exported Labels from KILI\n=========================\n\n".encode())
            fout.write(f"- Project name: {project_info['title']}\n".encode())
            fout.write(f"- Project identifier: {self.project_id}\n".encode())
            fout.write(f"- Project description: {project_info['description']}\n".encode())
            fout.write(f'- Export date: {datetime.now().strftime("%Y%m%d-%H%M%S")}\n'.encode())
            fout.write(f"- Exported format: {self.label_format}\n".encode())
            fout.write(f"- Exported labels: {self.export_type}\n".encode())

    @staticmethod
    def write_video_metadata_file(video_metadata: Dict, base_folder: Path) -> None:
        """
        Write video metadata file
        """
        video_metadata_json = json.dumps(video_metadata, sort_keys=True, indent=4)
        if video_metadata_json is not None:
            with (base_folder / "video_meta.json").open("wb") as output_file:
                output_file.write(video_metadata_json.encode("utf-8"))

    @staticmethod
    def write_remote_content_file(remote_content: List[str], images_folder: Path) -> None:
        """
        Write remote content file
        """
        remote_content_header = ["external id", "url", "label file"]
        # newline="" to disable universal newlines translation (bug fix for windows)
        with (images_folder / "remote_assets.csv").open("w", newline="", encoding="utf8") as file:
            writer = csv.writer(file)
            writer.writerow(remote_content_header)
            writer.writerows(remote_content)

    def export_project(
        self,
    ) -> None:
        """
        Export a project to a json.

        Return the name of the exported archive file.
        """
        self._check_arguments_compatibility()
        self._check_project_compatibility()

        self.logger.warning("Fetching assets...")

        with TemporaryDirectory() as export_root_folder:
            self.export_root_folder = export_root_folder
            assets = fetch_assets(
                self.auth,
                project_id=self.project_id,
                asset_ids=self.assets_ids,
                export_type=self.export_type,
                label_type_in=["DEFAULT", "REVIEW"],
                disable_tqdm=self.disable_tqdm,
                download_media=self.with_assets,
                local_media_dir=str(self.images_folder),
            )
            self.process_and_save(assets, self.output_file)

    @property
    def base_folder(self) -> Path:
        """
        Export base folder
        """
        return self.export_root_folder / self.project_id

    @property
    def images_folder(self) -> Path:
        """
        Export images folder
        """
        return self.base_folder / "images"

    @staticmethod
    def _filter_out_autosave_labels(assets: List[Asset]) -> List[Asset]:
        """
        Removes AUTOSAVE labels from exports
        """
        clean_assets = []
        for asset in assets:
            labels = asset.get("labels", [])
            clean_labels = list(filter(lambda label: label["labelType"] != "AUTOSAVE", labels))
            if clean_labels:
                asset["labels"] = clean_labels
            clean_assets.append(asset)
        return clean_assets

    @staticmethod
    def _format_json_response(label: Label, label_format: LabelFormat):
        """
        Format the label JSON response in the requested format
        """
        formatted_json_response = label.json_response(_format=label_format)
        json_response = {}
        for key, value in cast(Dict, formatted_json_response).items():
            if key.isdigit():
                json_response[int(key)] = value
                continue
            json_response[key] = value
        label["jsonResponse"] = json_response
        return label

    @staticmethod
    def pre_process_assets(assets: List[Asset], label_format: LabelFormat) -> List[Asset]:
        """
        Format labels in the requested format, and filter out autosave labels
        """
        assets_in_format = []
        for asset in assets:
            if "labels" in asset:
                labels_of_asset = []
                for label in asset["labels"]:
                    clean_label = AbstractExporter._format_json_response(label, label_format)
                    labels_of_asset.append(clean_label)
                asset["labels"] = labels_of_asset
            if "latestLabel" in asset:
                label = asset["latestLabel"]
                if label is not None:
                    clean_label = AbstractExporter._format_json_response(label, label_format)
                    asset["latestLabel"] = clean_label
            assets_in_format.append(asset)

        clean_assets = AbstractExporter._filter_out_autosave_labels(assets_in_format)
        return clean_assets



"""
Parser classes
"""
import csv
import json
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from kili.services.label_import.types import Classes


class AbstractLabelParser:  # pylint: disable=too-few-public-methods
    """
    Abstract label parser
    """

    def __init__(self, classes_by_id: Optional[Classes], target_job: Optional[str]) -> None:
        self.classes_by_id = classes_by_id
        self.target_job = target_job

    @abstractmethod
    def parse(self, label_file: Path) -> Dict[str, Any]:
        """
        Parses a label file
        """


class YoloLabelParser(AbstractLabelParser):  # pylint: disable=too-few-public-methods
    """
    Yolo label parser
    """

    def parse(self, label_file: Path) -> Dict[str, Any]:
        """
        Parses a Yolo label file
        """
        annotations = []
        assert self.classes_by_id
        assert self.target_job
        with open(label_file, "r", encoding="ascii") as l_f:
            csv_reader = csv.reader(l_f, delimiter=" ")
            for row in csv_reader:
                if len(row) == 0:
                    continue
                vertices, category, proba = self._parse(row)
                annotations.append(
                    {
                        "boundingPoly": [
                            {"normalizedVertices": [{"x": v[0], "y": v[1]} for v in vertices]}
                        ],
                        "categories": [
                            {
                                "name": self.classes_by_id[category],
                                "confidence": 100 if proba is None else int(100 * float(proba)),
                            }
                        ],
                    }
                )

        kili_json_response = {self.target_job: {"annotations": annotations}}

        return kili_json_response

    @staticmethod
    def _parse(row) -> Tuple[List[List[float]], int, Optional[float]]:
        try:
            class_id, x, y, width, height, proba = row  # pylint: disable=invalid-name
        except ValueError:
            class_id, x, y, width, height = row  # pylint: disable=invalid-name
            proba = None
        _class_id = int(class_id)
        half_width = float(width) / 2
        half_height = float(height) / 2
        _x = float(x)
        _y = float(y)
        return (
            [
                [_x - half_width, _y - half_height],
                [_x - half_width, _y + half_height],
                [_x + half_width, _y + half_height],
                [_x + half_width, _y - half_height],
            ],
            _class_id,
            proba,
        )


class KiliRawLabelParser(AbstractLabelParser):  # pylint: disable=too-few-public-methods
    """
    Kili raw label parser
    """

    def parse(self, label_file: Path) -> Dict[str, Any]:
        with label_file.open("r", encoding="utf-8") as l_f:
            return json.load(l_f)



"""
Types used by the conversion service
"""
from typing import Callable, Dict, NamedTuple, Tuple

from typing_extensions import Literal

ExportType = Literal["latest", "normal"]
SplitOption = Literal["split", "merged"]
LabelFormat = Literal["raw", "kili", "yolo_v4", "yolo_v5", "yolo_v7", "coco", "pascal_voc"]


class JobCategory(NamedTuple):
    """
    Contains information for a category
    """

    category_name: str
    id: int
    job_id: str


YoloAnnotation = Tuple[int, float, float, float, float]

CocoAnnotationModifier = Callable[[Dict, Dict, Dict], Dict]



"""
GraphQL Queries of Organizations
"""


from datetime import datetime
from typing import Dict, Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery
from kili.helpers import format_result


class OrganizationWhere(BaseQueryWhere):
    """
    Tuple to be passed to the OrganizationQuery to restrict query
    """

    def __init__(
        self,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
    ):
        self.email = email
        self.organization_id = organization_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK OrganizationWhere"""
        return {
            "id": self.organization_id,
            "user": {
                "email": self.email,
            },
        }


class OrganizationMetricsWhere(BaseQueryWhere):
    """
    Tuple to be passed to the OrganizationQuery to restrict query
    """

    def __init__(
        self,
        organization_id: str,
        start_date: datetime,
        end_date: datetime,
    ):
        self.organization_id = organization_id
        self.start_date = start_date
        self.end_date = end_date
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK OrganizationWhere"""
        return {
            "organizationId": self.organization_id,
            "startDate": self.start_date.isoformat(sep="T", timespec="milliseconds") + "Z",
            "endDate": self.end_date.isoformat(sep="T", timespec="milliseconds") + "Z",
        }


class OrganizationQuery(GraphQLQuery):
    """Organization query."""

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL organizations query
        """
        return f"""
        query organizations($where: OrganizationWhere!, $first: PageSize!, $skip: Int!) {{
            data: organizations(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countOrganizations($where: OrganizationWhere!) {
        data: countOrganizations(where: $where)
    }
    """

    GQL_ORGANIZATION_METRICS = """
    query organizationMetrics($where: OrganizationMetricsWhere!) {
        data: organizationMetrics(where: $where) {
            numberOfAnnotations
            numberOfHours
            numberOfLabeledAssets
        }
    }
    """

    def metrics(self, where: OrganizationMetricsWhere) -> Dict:
        """execute the organizationMetrics operation"""
        payload = {"where": where.graphql_payload}
        result = self.client.execute(self.GQL_ORGANIZATION_METRICS, payload)
        return format_result("data", result, None, self.ssl_verify)



"""
Common code for the Kili exporter.
"""

import json
import os
from pathlib import Path
from typing import List

from kili.orm import Asset
from kili.services.export.format.base import AbstractExporter

from ...media.video import cut_video


class KiliExporter(AbstractExporter):
    """
    Common code for Kili exporters.
    """

    ASSETS_DIR_NAME = "assets"

    def _check_arguments_compatibility(self):
        """
        Checks if the export label format is compatible with the export options.
        """

    def _check_project_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the project.
        """

    def _save_assets_export(self, assets: List[Asset], output_filename: Path):
        """
        Save the assets to a file and return the link to that file
        """
        self.logger.info("Exporting to kili format...")

        if self.with_assets:
            if self.project_input_type == "VIDEO":
                assets = self._cut_video_assets(assets)

            assets = self._clean_filepaths(assets)

        if self.single_file:
            project_json = json.dumps(assets, sort_keys=True, indent=4)
            self.base_folder.mkdir(parents=True, exist_ok=True)
            with (self.base_folder / "data.json").open("wb") as output_file:
                output_file.write(project_json.encode("utf-8"))
        else:
            labels_folder = self.base_folder / "labels"
            labels_folder.mkdir(parents=True, exist_ok=True)
            for asset in assets:
                external_id = asset["externalId"].replace(" ", "_")
                asset_json = json.dumps(asset, sort_keys=True, indent=4)
                with (labels_folder / f"{external_id}.json").open("wb") as output_file:
                    output_file.write(asset_json.encode("utf-8"))

        self.create_readme_kili_file(self.export_root_folder)

        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _clean_filepaths(self, assets: List[Asset]):
        """
        Remove TemporaryDirectory() prefix from filepaths in "jsonContent" and "content" fields.
        """
        for asset in assets:
            if os.path.isfile(asset["content"]):
                asset["content"] = str(Path(self.ASSETS_DIR_NAME) / Path(asset["content"]).name)

            json_content_list = []
            if isinstance(asset["jsonContent"], list):
                for filepath in asset["jsonContent"]:
                    if os.path.isfile(filepath):
                        json_content_list.append(
                            str(Path(self.ASSETS_DIR_NAME) / Path(filepath).name)
                        )
                asset["jsonContent"] = json_content_list
        return assets

    def _cut_video_assets(self, assets: List[Asset]):
        """
        Cut video assets into frames
        """
        for asset in assets:
            if asset["jsonContent"] == "" and os.path.isfile(asset["content"]):
                nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
                if nbr_frames == 0:
                    continue
                leading_zeros = len(str(nbr_frames))
                asset["jsonContent"] = cut_video(
                    video_path=asset["content"],
                    asset=asset,
                    leading_zeros=leading_zeros,
                    output_dir=self.images_folder,
                )
        return assets

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """
        Extract formatted annotations from labels and save the json in the buckets.
        """
        clean_assets = self.pre_process_assets(assets, self.label_format)
        return self._save_assets_export(
            clean_assets,
            output_filename,
        )

    @property
    def images_folder(self) -> Path:
        """
        Export images folder
        """
        return self.base_folder / self.ASSETS_DIR_NAME



"""CLI's project export subcommand"""

from typing import Optional, Union, cast

import click
from typeguard import typechecked
from typing_extensions import get_args

from kili import services
from kili.cli.common_args import Options
from kili.cli.helpers import get_kili_client
from kili.services.export.exceptions import NoCompatibleJobError
from kili.services.export.types import LabelFormat, SplitOption
from kili.services.types import ProjectId


@click.command(name="export")
@click.option(
    "--output-format",
    type=click.Choice(get_args(LabelFormat)),
    help="Format into which the label data will be converted",
    required=True,
)
@click.option(
    "--output-file", type=str, help="File into which the labels are saved.", required=True
)
@click.option(
    "--layout",
    type=click.Choice(get_args(SplitOption)),
    default="merged",
    help=(
        "Layout of the label files: 'split' to group labels per job, 'merged' to have one folder"
        " with every labels."
    ),
)
@click.option(
    "--single-file",
    type=bool,
    is_flag=True,
    help=(
        "Layout of the label files. Single file mode is only available for some specific formats"
        " (COCO and Kili)."
    ),
)
@click.option(
    "--with-assets/--without-assets",
    type=bool,
    default=True,
    help="Download assets in the export.",
)
@Options.api_key
@Options.endpoint
@Options.project_id
@Options.verbose
@Options.ssl_verify
@typechecked
# pylint: disable=too-many-arguments
def export_labels(
    output_format: LabelFormat,
    output_file: str,
    layout: SplitOption,
    single_file: bool,
    api_key: Optional[str],
    endpoint: Optional[str],
    project_id: str,
    verbose: bool,
    with_assets: bool,
    ssl_verify: Union[str, bool],
):
    """
    Export the Kili labels of a project to a given format.

    \b
    The supported formats are:

    - YOLO V4, V5, V7 for object detection tasks (bounding box).
    - Kili (a.k.a raw) for all tasks.
    - COCO for object detection tasks (semantic or bounding box)
    - Pascal VOC for object detection tasks.
    \b
    \b
    !!! warning "Cloud storage"
            Export is not allowed for projects connected to a cloud storage.
    \b
    \b
    !!! Examples
        ```
        kili project export \\
            --project-id <project_id> \\
            --output-format coco \\
            --output-file /tmp/export.zip
        ```
        ```
        kili project export \\
            --project-id <project_id> \\
            --output-format yolo_v5 \\
            --output-file /tmp/export_split.zip \\
            --layout split
        ```
    \b
    \b
    !!! warning "Unsupported exports"
        Currently, this command does not support the export of videos that have not
        been cut into separated frames.

        For such exports, please use the Kili UI.
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint, ssl_verify=ssl_verify)

    try:
        services.export_labels(
            kili.auth,
            asset_ids=None,
            project_id=cast(ProjectId, project_id),
            export_type="latest",
            label_format=output_format,
            split_option=layout,
            single_file=single_file,
            output_file=output_file,
            disable_tqdm=not verbose,
            log_level="INFO" if verbose else "WARNING",
            with_assets=with_assets,
            annotation_modifier=None,
        )
    except NoCompatibleJobError as excp:
        print(excp)



"""
Functions to import assets into a VIDEO_LEGACY project
"""
import mimetypes
import os
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from itertools import repeat
from typing import List

from kili.helpers import is_url
from kili.services.asset_import.base import (
    BaseAssetImporter,
    BaseBatchImporter,
    BatchParams,
    ContentBatchImporter,
    JsonContentBatchImporter,
)
from kili.services.asset_import.constants import (
    FRAME_IMPORT_BATCH_SIZE,
    IMPORT_BATCH_SIZE,
)
from kili.services.asset_import.exceptions import ImportValidationError
from kili.services.asset_import.types import AssetLike
from kili.utils import bucket


class VideoDataType(Enum):
    """
    Video data type
    """

    LOCAL_FRAMES = "LOCAL_FRAMES"
    HOSTED_FRAMES = "HOSTED_FRAMES"
    LOCAL_FILE = "LOCAL_FILE"
    HOSTED_FILE = "HOSTED_FILE"


class VideoMixin:
    """
    Helping functions for importing Video assets
    """

    @staticmethod
    def get_video_processing_parameters(asset: AssetLike, from_frames: bool):
        """
        Base method for adding video processing parameters
        """
        json_metadata = asset.get("json_metadata", {})
        processing_parameters = json_metadata.get("processingParameters", {})  # type: ignore X
        video_parameters = [
            ("shouldKeepNativeFrameRate", not from_frames),
            ("framesPlayedPerSecond", 30),
            ("shouldUseNativeVideo", not from_frames),
        ]
        for key, default_value in video_parameters:
            processing_parameters[key] = processing_parameters.get(key, default_value)
        return processing_parameters

    @staticmethod
    def map_frame_urls_to_index(asset: AssetLike):
        """
        Map a list of frame url to their index in the video
        """
        json_content = asset.get("json_content")
        assert json_content
        json_content_index = range(len(json_content))
        json_content = dict(zip(json_content_index, json_content))
        return AssetLike(**{**asset, "json_content": json_content})


class VideoContentBatchImporter(ContentBatchImporter, VideoMixin):
    """
    Class for importing a batch of video assets from content into a VIDEO project
    """

    def add_video_processing_parameters(self, asset):
        """
        Add video processing parameters for a content upload
        """
        json_metadata = asset.get("json_metadata", {})
        processing_parameters = self.get_video_processing_parameters(asset, from_frames=False)
        json_metadata = {**json_metadata, "processingParameters": processing_parameters}
        return AssetLike(**{**asset, "json_metadata": json_metadata})

    def import_batch(self, assets: List[AssetLike], verify: bool):
        """
        Import a batch of video assets from content into Kili.
        """
        assets = self.loop_on_batch(self.add_video_processing_parameters)(assets)
        return super().import_batch(assets, verify)


class FrameBatchImporter(JsonContentBatchImporter, VideoMixin):
    """
    Class for importing a batch of video assets from frames into a VIDEO project
    """

    def add_video_processing_parameters(self, asset):
        """
        Add video processing parameters for a frames upload
        """
        json_metadata = asset.get("json_metadata", {})
        processing_parameters = self.get_video_processing_parameters(asset, from_frames=True)
        json_metadata = {**json_metadata, "processingParameters": processing_parameters}
        return AssetLike(**{**asset, "json_metadata": json_metadata})

    def import_batch(self, assets: List[AssetLike], verify: bool):
        """
        Import a batch of video assets from frames
        """
        assets = self.add_ids(assets)
        if not self.is_hosted:
            assets = self.loop_on_batch(self.upload_frames_to_bucket)(assets)
        assets = self.loop_on_batch(self.map_frame_urls_to_index)(assets)
        assets = self.loop_on_batch(self.add_video_processing_parameters)(assets)
        return super().import_batch(assets, verify)

    def upload_frames_to_bucket(self, asset: AssetLike):
        """
        Import the local frames to the bucket
        """
        frames = asset.get("json_content")
        assert frames
        asset_id: str = asset.get("id") or f"unknown-{bucket.generate_unique_id()}"
        project_bucket_path = self.generate_project_bucket_path()
        asset_frames_paths = [
            BaseBatchImporter.build_url_from_parts(
                project_bucket_path, asset_id, "frame", str(frame_id)
            )
            for frame_id in range(len(frames))
        ]
        signed_urls = bucket.request_signed_urls(self.auth, asset_frames_paths)
        data_array = []
        content_type_array = []
        for frame_path in frames:
            with open(frame_path, "rb") as file:
                data_array.append(file.read())
            content_type, _ = mimetypes.guess_type(frame_path)
            content_type_array.append(content_type)
        with ThreadPoolExecutor() as threads:
            url_gen = threads.map(
                bucket.upload_data_via_rest,
                signed_urls,
                data_array,
                content_type_array,
                repeat(True),
            )
        cleaned_urls = (bucket.clean_signed_url(url, self.auth.api_endpoint) for url in url_gen)
        return AssetLike(**{**asset, "json_content": list(cleaned_urls)})


class VideoDataImporter(BaseAssetImporter):
    """
    Class for importing data into a VIDEO project
    """

    @staticmethod
    def is_hosted_frames(asset) -> bool:
        """
        Determine if the frames to import are hosted
        """
        frames = asset.get("json_content", [])
        return all(is_url(frame) for frame in frames)

    @staticmethod
    def has_any_local_frame(asset) -> bool:
        """
        Determine if at least one frame to import is a local file
        """
        frames = asset.get("json_content", [])
        return any(os.path.exists(frame) for frame in frames)

    def get_data_type(self, assets):
        """
        Determine the type of data to upload from the service payload
        """
        content_array = [asset.get("content", "") for asset in assets]
        has_json_content = any(asset.get("json_content") for asset in assets)
        if has_json_content:
            if any(content and not is_url(content) for content in content_array):
                raise ImportValidationError(
                    "Cannot import content that is not an url when importing a video from frames"
                )
            if all(self.is_hosted_frames(asset) for asset in assets):
                return VideoDataType.HOSTED_FRAMES
            return VideoDataType.LOCAL_FRAMES
        is_hosted_content = self.is_hosted_content(assets)
        if is_hosted_content:
            return VideoDataType.HOSTED_FILE
        return VideoDataType.LOCAL_FILE

    @staticmethod
    def should_cut_into_frames(assets) -> bool:
        """
        Determine if assets should be imported asynchronously and cut into frames
        """
        should_use_native_video_array = []
        for asset in assets:
            json_metadata = asset.get("json_metadata", {})
            processing_parameters = json_metadata.get("processingParameters", {})
            should_use_native_video_array.append(
                processing_parameters.get("shouldUseNativeVideo", True)
            )
        if all(should_use_native_video_array):
            return False
        if all(not b for b in should_use_native_video_array):
            return True
        raise ImportValidationError(
            """
            Cannot upload videos to split into frames
            and video to keep as native in the same time.
            Please separate the assets into 2 calls
            """
        )

    def import_assets(self, assets: List[AssetLike]):
        """
        Import video assets into Kili.
        """
        self._check_upload_is_allowed(assets)
        data_type = self.get_data_type(assets)
        assets = self.filter_duplicate_external_ids(assets)
        if data_type == VideoDataType.LOCAL_FILE:
            assets = self.filter_local_assets(assets, self.raise_error)
            as_frames = self.should_cut_into_frames(assets)
            batch_params = BatchParams(is_hosted=False, is_asynchronous=as_frames)
            batch_importer = VideoContentBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
            batch_size = IMPORT_BATCH_SIZE
        elif data_type == VideoDataType.HOSTED_FILE:
            as_frames = self.should_cut_into_frames(assets)
            batch_params = BatchParams(is_hosted=True, is_asynchronous=as_frames)
            batch_importer = VideoContentBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
            batch_size = IMPORT_BATCH_SIZE
        elif data_type == VideoDataType.LOCAL_FRAMES:
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = FrameBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
            batch_size = FRAME_IMPORT_BATCH_SIZE
        elif data_type == VideoDataType.HOSTED_FRAMES:
            batch_params = BatchParams(is_hosted=True, is_asynchronous=False)
            batch_importer = FrameBatchImporter(
                self.auth, self.project_params, batch_params, self.pbar
            )
            batch_size = IMPORT_BATCH_SIZE
        else:
            raise ImportValidationError
        result = self.import_assets_by_batch(assets, batch_importer, batch_size)
        return result



"""
GraphQL Queries of Plugins
"""


from datetime import datetime
from typing import List, NamedTuple, Optional

from kili.graphql import GraphQLQuery, QueryOptions
from kili.helpers import format_result


class PluginLogsWhere(NamedTuple):
    """
    Tuple to be passed to the PluginQuery to restrict query
    """

    project_id: str
    plugin_name: str
    start_date: Optional[datetime]


class PluginQuery(GraphQLQuery):
    """Plugin query."""

    @staticmethod
    def query(fragment):
        """The Plugin query is not implemented as other queries.
        To overpass the technical debt, the plugin query is access by calling the list method
        """
        _ = fragment
        return NotImplemented

    COUNT_QUERY = """
    query countPlugins($where: PluginWhere!) {
        data: countPlugins(where: $where)
    }
    """

    GQL_GET_PLUGIN_LOGS = """
    query getPluginLogs(
        $projectId: ID!
        $pluginName: String!
        $createdAt: DateTime
        $limit: Int
        $skip: Int
    ) {
        data: getPluginLogs(
            data: {
                projectId: $projectId
                pluginName: $pluginName
                createdAt: $createdAt
                limit: $limit
                skip: $skip
            }
        ) {
            content
            createdAt
            logType
            metadata {
                assetId
                labelId
            }
            pluginName
            projectId
            runId
        }
    }
    """

    @staticmethod
    def gql_list_plugins(fragment):
        """
        Return the GraphQL list_plugins query
        """
        return f"""
        query listPlugins{{
            data: listPlugins {{
                {fragment}
            }}
        }}
        """

    def list(self, fields: List[str]):
        """
        List plugins
        """
        fragment = self.fragment_builder(fields)
        query = self.gql_list_plugins(fragment)
        result = self.client.execute(query)
        return format_result("data", result, None, self.ssl_verify)

    def get_logs(self, where: PluginLogsWhere, options: QueryOptions):
        """
        Get logs of a plugin in Kili
        """

        payload = {
            "projectId": where.project_id,
            "pluginName": where.plugin_name,
            "limit": options.first,
            "skip": options.skip,
        }

        if where.start_date:
            payload["createdAt"] = (
                where.start_date.isoformat(sep="T", timespec="milliseconds") + "Z"
            )

        result = self.client.execute(self.GQL_GET_PLUGIN_LOGS, payload)
        return format_result("data", result, None, self.ssl_verify)


GQL_GET_PLUGIN_RUNNER_STATUS = """
    query getPluginRunnerStatus(
        $name: String!
    ) {
        data: getPluginRunnerStatus(
            name: $name
        )
    }
    """



"""
Services for data connections
"""
import logging
import time
from typing import Any, Dict, List, Optional

from tenacity import Retrying
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_exponential
from typing_extensions import Literal

from kili.graphql import QueryOptions
from kili.graphql.operations.asset.queries import AssetQuery, AssetWhere

from ...authentication import KiliAuth
from ...graphql.operations.data_connection.queries import (
    DataConnectionIdWhere,
    DataConnectionQuery,
)
from ...helpers import format_result
from ...mutations.data_connection.queries import (
    GQL_COMPUTE_DATA_CONNECTION_DIFFERENCES,
    GQL_VALIDATE_DATA_DIFFERENCES,
)

LOGGER = None


def _get_logger() -> logging.Logger:
    global LOGGER  # pylint: disable=global-statement

    if LOGGER is not None:
        return LOGGER

    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.INFO)
    LOGGER.addHandler(logging.StreamHandler())
    return LOGGER


def get_data_connection(auth: KiliAuth, data_connection_id: str, fields: List[str]) -> Dict:
    """
    Get data connection information
    """
    where = DataConnectionIdWhere(data_connection_id=data_connection_id)
    options = QueryOptions(first=1, disable_tqdm=True)
    data_connection = list(
        DataConnectionQuery(auth.client, auth.ssl_verify)(where, fields, options)
    )
    if len(data_connection) == 0:
        raise ValueError(f"No data connection with id {data_connection_id}")
    return data_connection[0]


def trigger_validate_data_differences(
    auth: KiliAuth, diff_type: Literal["ADD", "REMOVE"], data_connection_id: str
) -> Dict:
    """
    Call the validateDataDifferences resolver
    """
    variables = {
        "where": {"connectionId": data_connection_id, "type": diff_type},
        "processingParameters": None,
    }
    result = auth.client.execute(GQL_VALIDATE_DATA_DIFFERENCES, variables)
    data_connection = format_result("data", result, None, auth.ssl_verify)
    return data_connection


def validate_data_differences(
    auth: KiliAuth, diff_type: Literal["ADD", "REMOVE"], data_connection: Dict
) -> None:
    """
    Call the validateDataDifferences resolver and wait until the validation is done
    """
    diff = data_connection["dataDifferencesSummary"]["added" if diff_type == "ADD" else "removed"]

    where = AssetWhere(project_id=data_connection["projectId"])
    nb_assets_before = AssetQuery(auth.client, auth.ssl_verify).count(where)

    trigger_validate_data_differences(auth, diff_type, data_connection["id"])

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        stop=stop_after_delay(60),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            nb_assets_after = AssetQuery(auth.client, auth.ssl_verify).count(where)
            if abs(nb_assets_after - nb_assets_before) != diff:
                raise ValueError(
                    f"Number of assets after validation is not correct: before {nb_assets_before},"
                    f" after {nb_assets_after}, diff {diff}"
                )


def compute_differences(
    auth: KiliAuth, data_connection_id: str, blob_paths: Optional[List[str]] = None
) -> Dict:
    """
    Compute the data connection differences
    """
    variables: Dict[str, Any] = {"where": {"id": data_connection_id}}
    if blob_paths is not None:
        variables["data"] = {"blobPaths": blob_paths}
    result = auth.client.execute(GQL_COMPUTE_DATA_CONNECTION_DIFFERENCES, variables)
    data_connection = format_result("data", result, None, auth.ssl_verify)
    return data_connection


def verify_diff_computed(auth: KiliAuth, data_connection_id: str) -> None:
    """
    Verify that the data connection differences have been computed.

    Trigger the computation if not already computing.
    """
    logger = _get_logger()
    logger.info("Verifying that the data connection differences have been computed")

    data_connection = get_data_connection(auth, data_connection_id, fields=["isChecking"])

    if not data_connection["isChecking"]:
        compute_differences(auth, data_connection_id)
        time.sleep(1)  # backend needs some time...

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            data_connection = get_data_connection(auth, data_connection_id, fields=["isChecking"])
            if data_connection["isChecking"]:
                raise ValueError(f"Data connection is still checking: {data_connection}")

    time.sleep(1)  # backend needs some time...


def synchronize_data_connection(
    auth: KiliAuth, data_connection_id: str, delete_extraneous_files: bool
) -> Dict:
    """
    Launch a data connection synchronization
    """
    logger = _get_logger()
    logger.info("Synchronizing data connection: %s", data_connection_id)

    verify_diff_computed(auth, data_connection_id)

    data_connection = get_data_connection(
        auth,
        data_connection_id,
        fields=[
            "id",
            "dataDifferencesSummary.added",
            "dataDifferencesSummary.removed",
            "dataDifferencesSummary.total",
            "isChecking",
            "numberOfAssets",
            "projectId",
        ],
    )
    if data_connection["isChecking"]:
        raise ValueError(f"Data connection is still checking: {data_connection}")

    added = data_connection["dataDifferencesSummary"]["added"]
    removed = data_connection["dataDifferencesSummary"]["removed"]
    total = data_connection["dataDifferencesSummary"]["total"]
    nb_assets = data_connection["numberOfAssets"]

    logger.info("Currently %d asset(s) imported from the data connection.", nb_assets)

    if total == 0:
        logger.info("No differences found. Nothing to synchronize.")
        return data_connection

    logger.info(
        "Found %d difference(s): %d assets to add, %d assets to remove.", total, added, removed
    )

    if removed > 0:
        if delete_extraneous_files:
            validate_data_differences(auth, "REMOVE", data_connection)
            logger.info("Removed %d extraneous file(s).", removed)
        else:
            logger.info(
                "Use delete_extraneous_files=True to remove %d extraneous file(s).",
                removed,
            )

    if added > 0:
        validate_data_differences(auth, "ADD", data_connection)
        logger.info("Added %d file(s).", added)

    return get_data_connection(auth, data_connection_id, fields=["numberOfAssets", "projectId"])
